/*
 * This software is designed for the datalogger 'SGUduino 2024'. If you have a
 * different datalogger, then this software is most likely not working!
 *
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2023, Falko Schmidt
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <Arduino.h>
#include "Ulm_Weatherballoon.h"

// ====================================================================================================================

#if !defined(ARDUINO_UNOR4_MINIMA) && !defined(ARDUINO_UNOR4_WIFI)
#error This script only runs on Arduino UNO R4 boards!
#endif

// ====================================================================================================================

// --------------------------------------------------------------------------------------------------------------------
// SD CARD
// --------------------------------------------------------------------------------------------------------------------
constexpr uint8_t SD_CS_PIN = 10;
constexpr uint8_t SD_DETECT_PIN = 2;

// --------------------------------------------------------------------------------------------------------------------
// RGB LED (NeoPixel)
// --------------------------------------------------------------------------------------------------------------------
constexpr uint8_t LED_RGB_PIN = 9;

// --------------------------------------------------------------------------------------------------------------------
// MS5607 ALTITUDE SENSOR
// --------------------------------------------------------------------------------------------------------------------
constexpr uint8_t MS5611_ALTITUDE_SENSOR_I2C_ADDRESS = 0x77;

// --------------------------------------------------------------------------------------------------------------------
// DS18B20 TEMPERATURE SENSOR
// --------------------------------------------------------------------------------------------------------------------
constexpr uint8_t DS18B20_TEMPERATURE_SENSOR_PIN = 6;

// --------------------------------------------------------------------------------------------------------------------
// LSM6DS3 IMU (ACCELEROMETER AND GYROSCOPE)
// --------------------------------------------------------------------------------------------------------------------
constexpr uint8_t LSM6DS3_IMU_SENSOR_I2C_ADDRESS = 0x6A;

// --------------------------------------------------------------------------------------------------------------------
// INPUT DEVICES (SENSORS)
// --------------------------------------------------------------------------------------------------------------------
auto imu = Ulm_LSM6SDL(LSM6DS3_IMU_SENSOR_I2C_ADDRESS);
auto altitudeSensor = Ulm_MS5611(MS5611_ALTITUDE_SENSOR_I2C_ADDRESS);
auto temperatureSensor = Ulm_DS18B20(DS18B20_TEMPERATURE_SENSOR_PIN);

// --------------------------------------------------------------------------------------------------------------------
// OUTPUT DEVICES (LED AND DISPLAY)
// --------------------------------------------------------------------------------------------------------------------
auto builtInLed = Ulm_LED_BuiltIn();
auto led = Ulm_RGB_LED(LED_RGB_PIN);
auto display = Ulm_OLED_Display();

// --------------------------------------------------------------------------------------------------------------------
// STORAGE DEVICE
// --------------------------------------------------------------------------------------------------------------------
Ulm_SDStorage storage = Ulm_SDStorage::builder()
        .atPin(SD_CS_PIN)
        .withDetectPin(SD_DETECT_PIN)
        .atDirectory("SGUDUINO")
        .withDataFile("DAT_", Ulm_SDStorage::FileType::CSV)
        .withRedundancyFile("RED_", Ulm_SDStorage::FileType::CSV)
        .withLogicForExtremeEnvironments()
        .build();

// ====================================================================================================================
// DATA REPRESENTATION
// ====================================================================================================================
/**
 * Structure representing our raw sensor data.
 * All of these values will be sampled and stored to SD card.
 */
struct SensorData {
    unsigned long uptimeMillis;     // [MCU]        Uptime of the MCU (time since power on) in milliseconds
    float temperature;              // [DS18B20]    The ambient temperature in Celsius
    struct Ulm_LSM6SDL_Sample imuSensorSample;
    struct Ulm_MS5611_Sample altitudeSensorSample;
} sensorData;

/**
 * Storage routine for sensor data. The data will be formatted and then written to
 * SD card. After that, the onboard LED will be either green (success) or red (failure).
 */
void storeSensorData() {
    char buffer[128];
    snprintf(buffer, sizeof(buffer),
             "%lu;%.2f;%.2f;%.2f;%.2f;%.2f;%.2f;%.2f;%.2f;%.2f;%.2f\n",
             sensorData.uptimeMillis,
             sensorData.temperature,
             sensorData.imuSensorSample.accelerationX,
             sensorData.imuSensorSample.accelerationY,
             sensorData.imuSensorSample.accelerationZ,
             sensorData.imuSensorSample.gyroscopeX,
             sensorData.imuSensorSample.gyroscopeY,
             sensorData.imuSensorSample.gyroscopeZ,
             sensorData.altitudeSensorSample.temperature,
             sensorData.altitudeSensorSample.pressure
             //sensorData.altitudeSensorSample.pressure
    );
    storage.store(buffer) ? led.showSuccess() : led.showError();
}


void drawHeaderAndFooter() {
    display.setCursor(0, 0);
    display.write("SGUduino 2023/24");
    display.drawLine(0, 8, display.width(), 8, SSD1306_WHITE);
    display.drawLine(0, 54, display.width(), 54, SSD1306_WHITE);
    display.setCursor(0, 56);
    display.write("Falko Schmidt, UULM");
}


void showSensorData() {
    display.clearDisplay();
    drawHeaderAndFooter();

    const unsigned int min = sensorData.uptimeMillis / 60000;
    const unsigned int s = sensorData.uptimeMillis / 1000 % 60;
    const unsigned int ms = sensorData.uptimeMillis % 1000;

    display.setCursor(0, 10);
    display.write("UPTIME:");
    display.print(min < 10 ? "0" : "");
    display.print(min);
    display.print(':');
    display.print(s < 10 ? "0" : "");
    display.print(s);
    display.print(':');
    display.print(ms < 100 ? "0" : "");
    display.print(ms < 10 ? "0" : "");
    display.print(ms);

    display.setCursor(0, 19);
    display.write(" ACCEL:");
    display.print(sensorData.imuSensorSample.accelerationX, 1);
    display.write(' ');
    display.print(sensorData.imuSensorSample.accelerationY, 1);
    display.write(' ');
    display.print(sensorData.imuSensorSample.accelerationZ, 1);

    display.setCursor(0, 28);
    display.write("  GYRO:");
    display.print(sensorData.imuSensorSample.gyroscopeX, 0);
    display.write(' ');
    display.print(sensorData.imuSensorSample.gyroscopeY, 0);
    display.write(' ');
    display.print(sensorData.imuSensorSample.gyroscopeZ, 0);

    display.setCursor(0, 37);
    display.write("  TEMP:");
    display.print(sensorData.temperature, 1);
    display.write('C');

    display.setCursor(0, 46);
    display.write(" PRESS:");
    display.print(sensorData.altitudeSensorSample.pressure, 2);
    display.write("mBar");

    display.display();
}


/**
 * Depending on the value of 'success', the display will be updated
 * providing information, if a module was inited successfully or not.
 * @param success whether the module was inited successfully or not.
 * @param moduleName the name of the module, that was inited.
 * @return the value of 'success'.
 */
bool initAndDisplay(bool success, const char *moduleName) {
    display.write('[');
    display.write(success ? " OK " : "FAIL");
    display.write("] ");
    display.println(moduleName);
    return success;
}

// ====================================================================================================================
// PROGRAM ROUTINE
// ====================================================================================================================
/**
 * This function runs once and initializes all interfaces, sensors and actors of the implemented datalogger.
 * Also, a headline will be written into the datafile.
 */
void setup() {

    // No need to check return value
    // for built-in LED.
    builtInLed.begin();

    bool hasErrors = false;

    // Start LED first...
    led.begin();
    led.showWarning();

    // ...then start display...
    if (display.begin()) {
        drawHeaderAndFooter();
        display.display();
    }

    // Now we can start sensors one by one, starting at the top of the PCB
    display.setCursor(0, 16);
    hasErrors |= initAndDisplay(temperatureSensor.begin(), "Thermometer");

    // IMU (LSM6DS3 via i2c)
    display.setCursor(0, 24);
    hasErrors |= initAndDisplay(imu.begin(), "IMU");

    // Altimeter (MS5607 via i2c)
    display.setCursor(0, 32);
    hasErrors |= initAndDisplay(altitudeSensor.begin(), "Altimeter");

    // And finally: the SD card!
    display.setCursor(0, 40);
    hasErrors |= initAndDisplay(storage.begin(), "microSD");
    display.display();

    // Write header line of data.
    String header = "";
    header += "Uptime [ms];";
    header += "Temperature [C];";
    header += "Acceleration X [g];";
    header += "Acceleration Y [g];";
    header += "Acceleration Z [g];";
    header += "Gyroscope X [deg/s];";
    header += "Gyroscope Y [deg/s];";
    header += "Gyroscope Z [deg/s];";
    header += "Temperature (Alt.sensor) [C]";
    header += "Pressure [mBar];";
    header += "Altitude [m]";
    header += '\n';

    hasErrors |= storage.store(header);
    hasErrors ? builtInLed.on() : builtInLed.off();
    hasErrors ? led.showError() : led.showSuccess();

    // Just make sure, that user can see the error messages.
    delay(5000);
}

/**
 * This function runs forever.
 * We will use it to implement our sampling and storage routine.
 */
void loop() {

    // Get the start time, which will be used for accurate waiting.
    unsigned long startTime = millis();

    // Current uptime in milliseconds.
    sensorData.uptimeMillis = startTime;

    // Sample all sensors
    bool sampleSuccess = imu.sample(sensorData.imuSensorSample);
    sampleSuccess &= altitudeSensor.sample(sensorData.altitudeSensorSample);
    sampleSuccess &= temperatureSensor.sample(sensorData.temperature);

    // Check, if there was an error
    sampleSuccess ? led.showSuccess() : led.showError();

    // Store data.
    storeSensorData();

    // Show the current values on screen.
    showSensorData();

    // Wait for one second.
    unsigned long runTime = millis() - startTime;
    if(runTime < 1000) {
        // Only wait, if the needed time for sampling and storage
        // was less than one second. Otherwise, we will not add
        // additional waiting time.
        delay(1000 - runTime);
    }
}
